#!/bin/python

import itertools
import math
import operator
import os
import random
import sys

import numpy as np
import numpy.random as rnd
import scipy.cluster.hierarchy as sch
import scipy.io as sio

def linkage_to_ordering(Z):
    n = len(Z) + 1
    cache = dict()
    for k in range(len(Z)):
        c1, c2 = int(Z[k][0]), int(Z[k][1])
        c1 = [c1] if c1 < n else cache.pop(c1)
        c2 = [c2] if c2 < n else cache.pop(c2)
        cache[n+k] = c1 + c2
    return cache[2*len(Z)]

def corrcov(C):
    '''
    Convert covariance matrix to correlation matrix
    '''
    d = np.diag(C)
    scaling = np.sqrt(np.outer(d,d))
    return np.divide(C, scaling)

def cluster_graph(K, c, m_within, singletons):
    '''
    K: num variables
    c: num non-singleton clusters
    n0: initial size of non-singleton clusters
    m_within: average degree of non-singleton variables
    singletons: number of singleton variables (must be <= K - 2*c)
    '''
    # Initializes network
    graph = np.zeros((K,K))
    n0 = 2
    commun = {}
    for i in range(0, c):
        graph[i*n0:(i+1)*n0,i*n0:(i+1)*n0] = np.ones((n0,n0)) - np.eye(n0)
        commun[i] = range(i*n0,(i+1)*n0)

    # Grows network
    (within_frac, within_int) = np.modf(m_within)
    num_within = within_int + rnd.binomial(1, within_frac, size=K)
    rnd_commun = rnd.randint(0, c, size=K) 
    for i in range(c*n0, K - singletons):
        deg_dist = np.sum(graph,0)
        cur_c = rnd_commun[i] # My community index
        fellows = commun[cur_c]

        deg_dist_within = deg_dist[fellows] / np.sum(deg_dist[fellows])
        chosen_within = rnd.choice(fellows,
            size=min(num_within[i],len(fellows)), replace=False, p=deg_dist_within)
        graph[i, chosen_within] = 1
        graph[chosen_within, i] = 1
        
        commun[cur_c].append(i)

    # Reorders nodes in network
    dummyPrec = np.diagflat(np.sum(graph,0)+0.1) - graph
    Cov = np.linalg.inv(dummyPrec)
    Corr = corrcov(Cov)
    link = sch.linkage(Corr, method="average", metric="correlation")
    ix = linkage_to_ordering(link)
    graph = graph[:,ix]
    graph = graph[ix,:]

    return graph

if __name__ == "__main__":
    if len(sys.argv) < 4 or len(sys.argv) > 5:
        print ("usage: numVariables numClusters numSingletons [avgDegreeWithinCluster=4]")
        exit(0)
    K = int(sys.argv[1])
    c = int(sys.argv[2])
    singletons = int(sys.argv[3])
    within = 4
    if len(sys.argv) == 5:
        within = float(sys.argv[4])
    epsilon = 0.5 # minimum eigenvalue of precision matrix
    graph = cluster_graph(K, c, within, singletons)
    # Precision matrix from graph laplacian:
    Prec = np.diagflat(np.sum(graph,1)+epsilon) - graph
    sio.savemat('glasso-data.mat', {'Prec' : Prec}, oned_as='column')

