function [fdrF, fdrS, fdrT, pF, pS, pT] = glasso(n, alpha, S)
    close all;
    format long;
    set(0,'defaulttextinterpreter','latex');
    load('glasso-data.mat', 'Prec');
    d = size(Prec, 1);
    if nargin < 3
        trueCov = inv(Prec);
        data = mvnrnd(zeros(1,d),trueCov, n);
        S = cov(data);
    end
    
    [num_true, graph_true] = graphconncomp(sparse(Prec),'Directed', false);
    S_offdiag = abs(triu(S,1));
    S_offdiag = S_offdiag(:);
    S_offdiag_sort = sort(unique(S_offdiag),'descend');
    maxAbsEmpCov = S_offdiag_sort(1);
    minAbsEmpCov = S_offdiag_sort(end);
    
    num_points = length(S_offdiag_sort);
    testpoints = S_offdiag_sort;
    num_components = zeros(num_points,1);  
    for t_ix=1:num_points
        t = testpoints(t_ix);
        S_thresh = sparse(abs(S) > t);
        [numcon, conarray] = graphconncomp(S_thresh,'Directed', false);
        num_components(t_ix) = numcon;
    end
    
    twiddle_rho = [];
    twiddle_components = [];
    for t_ix=1:num_points-1
        if num_components(t_ix) ~= num_components(t_ix+1)
            twiddle_rho = [twiddle_rho; testpoints(t_ix)];
            twiddle_components = [twiddle_components; num_components(t_ix)];
        end
    end
    num_merges = length(twiddle_rho);
    is_correct_merge = zeros(num_merges-1,1);
    for t_ix=1:num_merges-1
        S_thresh_before = sparse(abs(S) > twiddle_rho(t_ix)+1e-7);
        S_thresh_after = sparse(abs(S) > twiddle_rho(t_ix)-1e-7);
        [i, j] = ind2sub([d, d], find((~S_thresh_before) & S_thresh_after));
        if graph_true(i(1))==graph_true(j(1))
            is_correct_merge(t_ix) = 1;
        end
    end
    
    T_old = n*twiddle_rho(1:end-1).*(-1*diff(twiddle_rho));
    pvalues_old = zeros(size(T_old));
    for i=1:size(T_old,1)
        pvalues_old(i) = 1 - expcdf(T_old(i),1);
    end
    
    f = @(x) (2*(1-x.^2).^((n-4)/2)/beta(0.5,(n-2)/2));
    Fbar = @(x) integral(f, x, 1);
    T_func = @(lk,lkk) log(Fbar(lkk))-log(Fbar(lk));
    num_merges_1 = length(twiddle_rho) -1;
    T = zeros(num_merges_1,1);
    for i=1:num_merges_1
        T(i) = T_func(twiddle_rho(i), twiddle_rho(i+1));
        if isnan(T(i))
            T(i) = T_old(i);
        end
    end
    pvalues = zeros(num_merges_1, 1);
    for i=1:num_merges_1
        pvalues(i) = 1 - expcdf(T(i),1);
    end  
    kF = forwardstop(pvalues, alpha);
    kS = strongstop(pvalues, alpha);
    kT = tailstop(T, alpha);
    
    fdrF = sum(~is_correct_merge(1:kF))/kF;
    if kF==0
        fdrF = 0;
    end
    fdrS = sum(~is_correct_merge(1:kS))/kS;
    if kS==0
        fdrS = 0;
    end
    fdrT = sum(~is_correct_merge(1:kT))/kT;
    if kT==0
        fdrT = 0;
    end
    pF = sum(is_correct_merge(1:kF))/sum(is_correct_merge);
    pS = sum(is_correct_merge(1:kS))/sum(is_correct_merge);
    pT = sum(is_correct_merge(1:kT))/sum(is_correct_merge);
    
    if nargin == 3
        return;
    end

    figure('name', 'Inputs'); 
    subplot(1,3,1); imagesc(Prec); title('true precision matrix');
    subplot(1,3,2); imagesc(trueCov); title('true covariance matrix');
    subplot(1,3,3); imagesc(S); title('empirical covariance matrix');    
    xlabel(sprintf('minabs=%3.2f maxabs=%3.2f',minAbsEmpCov,maxAbsEmpCov));
    
    
    figure('name', 'regularization path'); 
    subplot(1,3,1);
    plot(testpoints, num_components); 
    xlabel('rho'); ylabel('num components');
    title('components');
    subplot(1,3,2);
    plot((1:length(twiddle_rho))', twiddle_rho);
    xlabel('k'); ylabel('rho');
    title('location of knot points');
    subplot(1,3,3);
    plot(twiddle_components, twiddle_rho);
    xlabel('num components'); ylabel('rho');
    figure('name','hypothesis testing');    
    merge_ix = (1:length(pvalues))';
    ymin = 0; ymax = 1;
    plot(merge_ix(is_correct_merge==1), pvalues(is_correct_merge==1),'+', ...
        merge_ix(is_correct_merge==0), pvalues(is_correct_merge==0), 'x', ...
        [kF, kF],[ymin ymax],'-', ...
        [kS, kS],[ymin ymax],'--', ...
        [kT, kT], [ymin ymax],'-.'); 
    xlabel('merge number');
    ylabel('p-value');
    title(['$$\alpha=' sprintf('%3.2f, n=%i',alpha,n) '$$']);
    legend('correct', 'incorrect','ForwardStop','StrongStop','TailStop');
end



function [kF] = forwardstop(pvalues, alpha)
    % Rejects the first kF hypotheses
    y = -1*log(1 - pvalues);
    y_cumsum = cumsum(y);
    q = y_cumsum ./ (1:length(y_cumsum))';
    kF = max(find(q <= alpha));
    if isempty(kF)
        kF = 0;
    end
end

function [kS] = strongstop(pvalues, alpha)
    y = -1*log(pvalues);
    m = length(y);
    yj_j = y ./ (1:m)';
    z = flipud(cumsum(flipud(yj_j)));
    q = exp(-1*z);
    q_k = q ./ (1:length(q))';
    
    kS = max(find(q_k < alpha/m));
    if isempty(kS)
        kS = 0;
    end
end

function [kT] = tailstop(T, alpha)
    m = length(T);
    tau = 10;
    q = zeros(m, 1);
    for i=1:m
        factor = min((i:m)',tau) ./ (i:m)';
        arr = T(i:m) .* factor;
        q(i) = exp(-1*sum(arr));
    end
    kT = 0;
    for k=1:m
        if q(k) < alpha*k/m
            kT = k;
        end
    end
end


